@@ -19,46 +19,44 @@ processing but threads to do work.
 import re
 from io import BytesIO
 
-from waitress.compat import (
-    tostr,
-    urlparse,
-    unquote_bytes_to_wsgi,
-)
-
 from waitress.buffers import OverflowableBuffer
-
-from waitress.receiver import (
-    FixedStreamReceiver,
-    ChunkedReceiver,
-)
-
+from waitress.compat import tostr, unquote_bytes_to_wsgi, urlparse
+from waitress.receiver import ChunkedReceiver, FixedStreamReceiver
 from waitress.utilities import (
-    find_double_newline,
+    BadRequest,
     RequestEntityTooLarge,
     RequestHeaderFieldsTooLarge,
-    BadRequest,
+    ServerNotImplemented,
+    find_double_newline,
 )
+from .rfc7230 import HEADER_FIELD
+
 
 class ParsingError(Exception):
     pass
 
+
+class TransferEncodingNotImplemented(Exception):
+    pass
+
 class HTTPRequestParser(object):
     """A structure that collects the HTTP request.
 
     Once the stream is completed, the instance is passed to
     a server task constructor.
     """
-    completed = False        # Set once request is completed.
-    empty = False            # Set if no request was made.
+
+    completed = False  # Set once request is completed.
+    empty = False  # Set if no request was made.
     expect_continue = False  # client sent "Expect: 100-continue" header
-    headers_finished = False # True when headers have been read
-    header_plus = b''
+    headers_finished = False  # True when headers have been read
+    header_plus = b""
     chunked = False
     content_length = 0
     header_bytes_received = 0
     body_bytes_received = 0
     body_rcv = None
-    version = '1.0'
+    version = "1.0"
     error = None
     connection_close = False
 
@@ -81,19 +79,49 @@ class HTTPRequestParser(object):
         body have been received.
         """
         if self.completed:
-            return 0 # Can't consume any more.
+            return 0  # Can't consume any more.
+
         datalen = len(data)
         br = self.body_rcv
         if br is None:
             # In header.
+            max_header = self.adj.max_request_header_size
+
             s = self.header_plus + data
             index = find_double_newline(s)
+            consumed = 0
+
+            if index >= 0:
+                # If the headers have ended, and we also have part of the body
+                # message in data we still want to validate we aren't going
+                # over our limit for received headers.
+                self.header_bytes_received += index
+                consumed = datalen - (len(s) - index)
+            else:
+                self.header_bytes_received += datalen
+                consumed = datalen
+
+            # If the first line + headers is over the max length, we return a
+            # RequestHeaderFieldsTooLarge error rather than continuing to
+            # attempt to parse the headers.
+            if self.header_bytes_received >= max_header:
+                self.parse_header(b"GET / HTTP/1.0\r\n")
+                self.error = RequestHeaderFieldsTooLarge(
+                    "exceeds max_header of %s" % max_header
+                )
+                self.completed = True
+                return consumed
+
             if index >= 0:
                 # Header finished.
                 header_plus = s[:index]
-                consumed = len(data) - (len(s) - index)
-                # Remove preceeding blank lines.
+
+                # Remove preceeding blank lines. This is suggested by
+                # https://tools.ietf.org/html/rfc7230#section-3.5 to support
+                # clients sending an extra CR LF after another request when
+                # using HTTP pipelining
                 header_plus = header_plus.lstrip()
+
                 if not header_plus:
                     self.empty = True
                     self.completed = True
@@ -103,45 +131,42 @@ class HTTPRequestParser(object):
                     except ParsingError as e:
                         self.error = BadRequest(e.args[0])
                         self.completed = True
+                    except TransferEncodingNotImplemented as e:
+                        self.error = ServerNotImplemented(e.args[0])
+                        self.completed = True
                     else:
                         if self.body_rcv is None:
                             # no content-length header and not a t-e: chunked
                             # request
                             self.completed = True
+
                         if self.content_length > 0:
                             max_body = self.adj.max_request_body_size
                             # we won't accept this request if the content-length
                             # is too large
+
                             if self.content_length >= max_body:
                                 self.error = RequestEntityTooLarge(
-                                    'exceeds max_body of %s' % max_body)
+                                    "exceeds max_body of %s" % max_body
+                                )
                                 self.completed = True
                 self.headers_finished = True
+
                 return consumed
-            else:
-                # Header not finished yet.
-                self.header_bytes_received += datalen
-                max_header = self.adj.max_request_header_size
-                if self.header_bytes_received >= max_header:
-                    # malformed header, we need to construct some request
-                    # on our own. we disregard the incoming(?) requests HTTP
-                    # version and just use 1.0. IOW someone just sent garbage
-                    # over the wire
-                    self.parse_header(b'GET / HTTP/1.0\n')
-                    self.error = RequestHeaderFieldsTooLarge(
-                        'exceeds max_header of %s' % max_header)
-                    self.completed = True
-                self.header_plus = s
-                return datalen
+
+            # Header not finished yet.
+            self.header_plus = s
+
+            return datalen
         else:
             # In body.
             consumed = br.received(data)
             self.body_bytes_received += consumed
             max_body = self.adj.max_request_body_size
+
             if self.body_bytes_received >= max_body:
                 # this will only be raised during t-e: chunked requests
-                self.error = RequestEntityTooLarge(
-                    'exceeds max_body of %s' % max_body)
+                self.error = RequestEntityTooLarge("exceeds max_body of %s" % max_body)
                 self.completed = True
             elif br.error:
                 # garbage in chunked encoding input probably
@@ -150,6 +175,7 @@ class HTTPRequestParser(object):
             elif br.completed:
                 # The request (with the body) is ready to use.
                 self.completed = True
+
                 if self.chunked:
                     # We've converted the chunked transfer encoding request
                     # body into a normal request body, so we know its content
@@ -157,7 +183,8 @@ class HTTPRequestParser(object):
                     # TRANSFER_ENCODING header in parse_header, so this will
                     # appear to the client to be an entirely non-chunked HTTP
                     # request with a valid content-length.
-                    self.headers['CONTENT_LENGTH'] = str(br.__len__())
+                    self.headers["CONTENT_LENGTH"] = str(br.__len__())
+
             return consumed
 
     def parse_header(self, header_plus):
@@ -165,36 +192,43 @@ class HTTPRequestParser(object):
         Parses the header_plus block of text (the headers plus the
         first line of the request).
         """
-        index = header_plus.find(b'\n')
+        index = header_plus.find(b"\r\n")
         if index >= 0:
             first_line = header_plus[:index].rstrip()
-            header = header_plus[index + 1:]
+            header = header_plus[index + 2 :]
         else:
-            first_line = header_plus.rstrip()
-            header = b''
+            raise ParsingError("HTTP message header invalid")
 
-        self.first_line = first_line # for testing
+        if b"\r" in first_line or b"\n" in first_line:
+            raise ParsingError("Bare CR or LF found in HTTP message")
+
+        self.first_line = first_line  # for testing
 
         lines = get_header_lines(header)
 
         headers = self.headers
         for line in lines:
-            index = line.find(b':')
-            if index > 0:
-                key = line[:index]
-                if b'_' in key:
-                    continue
-                value = line[index + 1:].strip()
-                key1 = tostr(key.upper().replace(b'-', b'_'))
-                # If a header already exists, we append subsequent values
-                # seperated by a comma. Applications already need to handle
-                # the comma seperated values, as HTTP front ends might do
-                # the concatenation for you (behavior specified in RFC2616).
-                try:
-                    headers[key1] += tostr(b', ' + value)
-                except KeyError:
-                    headers[key1] = tostr(value)
-            # else there's garbage in the headers?
+            header = HEADER_FIELD.match(line)
+
+            if not header:
+                raise ParsingError("Invalid header")
+
+            key, value = header.group('name', 'value')
+
+            if b"_" in key:
+                # TODO(xistence): Should we drop this request instead?
+                continue
+
+            value = value.strip()
+            key1 = tostr(key.upper().replace(b"-", b"_"))
+            # If a header already exists, we append subsequent values
+            # seperated by a comma. Applications already need to handle
+            # the comma seperated values, as HTTP front ends might do
+            # the concatenation for you (behavior specified in RFC2616).
+            try:
+                headers[key1] += tostr(b", " + value)
+            except KeyError:
+                headers[key1] = tostr(value)
 
         # command, uri, version will be bytes
         command, uri, version = crack_first_line(first_line)
@@ -202,37 +236,62 @@ class HTTPRequestParser(object):
         command = tostr(command)
         self.command = command
         self.version = version
-        (self.proxy_scheme,
-         self.proxy_netloc,
-         self.path,
-         self.query, self.fragment) = split_uri(uri)
+        (
+            self.proxy_scheme,
+            self.proxy_netloc,
+            self.path,
+            self.query,
+            self.fragment,
+        ) = split_uri(uri)
         self.url_scheme = self.adj.url_scheme
-        connection = headers.get('CONNECTION', '')
+        connection = headers.get("CONNECTION", "")
 
-        if version == '1.0':
-            if connection.lower() != 'keep-alive':
+        if version == "1.0":
+            if connection.lower() != "keep-alive":
                 self.connection_close = True
 
-        if version == '1.1':
+        if version == "1.1":
             # since the server buffers data from chunked transfers and clients
             # never need to deal with chunked requests, downstream clients
             # should not see the HTTP_TRANSFER_ENCODING header; we pop it
             # here
-            te = headers.pop('TRANSFER_ENCODING', '')
-            if te.lower() == 'chunked':
+            te = headers.pop("TRANSFER_ENCODING", "")
+
+            encodings = [encoding.strip().lower() for encoding in te.split(",") if encoding]
+
+            for encoding in encodings:
+                # Out of the transfer-codings listed in
+                # https://tools.ietf.org/html/rfc7230#section-4 we only support
+                # chunked at this time.
+
+                # Note: the identity transfer-coding was removed in RFC7230:
+                # https://tools.ietf.org/html/rfc7230#appendix-A.2 and is thus
+                # not supported
+                if encoding not in {"chunked"}:
+                    raise TransferEncodingNotImplemented(
+                        "Transfer-Encoding requested is not supported."
+                    )
+
+            if encodings and encodings[-1] == "chunked":
                 self.chunked = True
                 buf = OverflowableBuffer(self.adj.inbuf_overflow)
                 self.body_rcv = ChunkedReceiver(buf)
-            expect = headers.get('EXPECT', '').lower()
-            self.expect_continue = expect == '100-continue'
-            if connection.lower() == 'close':
+            elif encodings:  # pragma: nocover
+                raise TransferEncodingNotImplemented(
+                    "Transfer-Encoding requested is not supported."
+                )
+
+            expect = headers.get("EXPECT", "").lower()
+            self.expect_continue = expect == "100-continue"
+            if connection.lower() == "close":
                 self.connection_close = True
 
         if not self.chunked:
             try:
-                cl = int(headers.get('CONTENT_LENGTH', 0))
+                cl = int(headers.get("CONTENT_LENGTH", 0))
             except ValueError:
-                cl = 0
+                raise ParsingError("Content-Length is invalid")
+
             self.content_length = cl
             if cl > 0:
                 buf = OverflowableBuffer(self.adj.inbuf_overflow)
@@ -250,10 +309,34 @@ class HTTPRequestParser(object):
         if body_rcv is not None:
             body_rcv.getbuf().close()
 
+
 def split_uri(uri):
     # urlsplit handles byte input by returning bytes on py3, so
     # scheme, netloc, path, query, and fragment are bytes
-    scheme, netloc, path, query, fragment = urlparse.urlsplit(uri)
+
+    scheme = netloc = path = query = fragment = b""
+
+    # urlsplit below will treat this as a scheme-less netloc, thereby losing
+    # the original intent of the request. Here we shamelessly stole 4 lines of
+    # code from the CPython stdlib to parse out the fragment and query but
+    # leave the path alone. See
+    # https://github.com/python/cpython/blob/8c9e9b0cd5b24dfbf1424d1f253d02de80e8f5ef/Lib/urllib/parse.py#L465-L468
+    # and https://github.com/Pylons/waitress/issues/260
+
+    if uri[:2] == b"//":
+        path = uri
+
+        if b"#" in path:
+            path, fragment = path.split(b"#", 1)
+
+        if b"?" in path:
+            path, query = path.split(b"?", 1)
+    else:
+        try:
+            scheme, netloc, path, query, fragment = urlparse.urlsplit(uri)
+        except UnicodeError:
+            raise ParsingError("Bad URI")
+
     return (
         tostr(scheme),
         tostr(netloc),
@@ -262,37 +345,58 @@ def split_uri(uri):
         tostr(fragment),
     )
 
+
 def get_header_lines(header):
     """
     Splits the header into lines, putting multi-line headers together.
     """
     r = []
-    lines = header.split(b'\n')
+    lines = header.split(b"\r\n")
     for line in lines:
-        if line.startswith((b' ', b'\t')):
+        if not line:
+            continue
+
+        if b"\r" in line or b"\n" in line:
+            raise ParsingError('Bare CR or LF found in header line "%s"' % tostr(line))
+
+        if line.startswith((b" ", b"\t")):
             if not r:
-                # http://corte.si/posts/code/pathod/pythonservers/index.html
+                # https://corte.si/posts/code/pathod/pythonservers/index.html
                 raise ParsingError('Malformed header line "%s"' % tostr(line))
             r[-1] += line
         else:
             r.append(line)
     return r
 
+
 first_line_re = re.compile(
-    b'([^ ]+) '
-    b'((?:[^ :?#]+://[^ ?#/]*(?:[0-9]{1,5})?)?[^ ]+)'
-    b'(( HTTP/([0-9.]+))$|$)'
+    b"([^ ]+) "
+    b"((?:[^ :?#]+://[^ ?#/]*(?:[0-9]{1,5})?)?[^ ]+)"
+    b"(( HTTP/([0-9.]+))$|$)"
 )
 
+
 def crack_first_line(line):
     m = first_line_re.match(line)
     if m is not None and m.end() == len(line):
         if m.group(3):
             version = m.group(5)
         else:
-            version = None
-        command = m.group(1).upper()
+            version = b""
+        method = m.group(1)
+
+        # the request methods that are currently defined are all uppercase:
+        # https://www.iana.org/assignments/http-methods/http-methods.xhtml and
+        # the request method is case sensitive according to
+        # https://tools.ietf.org/html/rfc7231#section-4.1
+
+        # By disallowing anything but uppercase methods we save poor
+        # unsuspecting souls from sending lowercase HTTP methods to waitress
+        # and having the request complete, while servers like nginx drop the
+        # request onto the floor.
+        if method != method.upper():
+            raise ParsingError('Malformed HTTP method "%s"' % tostr(method))
         uri = m.group(2)
-        return command, uri, version
+        return method, uri, version
     else:
-        return b'', b'', b''
+        return b"", b"", b""
